require  'forwardable'
require 'csv'
require 'parseexcel'

module Hyogen
  class Reader
    module Tag
      VTABLE = /\{VTABLE\}/
      HTABLE = /\{HTABLE\}/
      LABELS = /\{L\}/
      LABEL_NOTE = /\{N\}/
      ALIAS = /<(.+)>/
      
      def self.aliased(str)
        result = remove_space( (str =~ Tag::ALIAS) ? $1 : str )
        (result || "").sub(LABEL_NOTE,"")
      end

      def self.lookup_tag(obj, tag)
        return nil  if obj == nil
        list = (obj.kind_of?(Array) ? obj : [obj])
        data = nil
        position = nil
        matched_value = nil
        list.each_with_index do |value, i|
          if tag =~ value
            matched_value = value
            position = i
            m = Regexp.last_match
            data = remove_space(m.pre_match)
            data = remove_space(m.post_match)  if data == ""
            break
          end
        end
        return nil  unless matched_value
        {:value => matched_value, :tag => tag, :data => data, :position => position}
      end

      def self.lookup_table_tag(obj)
        lookup_tag(obj, Tag::VTABLE)  or lookup_tag(obj, Tag::HTABLE)
      end
      
      private
      def self.remove_space(str)
        return nil  unless str
        str.sub(/^\s*/,"").sub(/\s*$/,"")
      end
    end
    
    class Matrix
      extend Forwardable
      def initialize data
        @data = data #񎟌z
      end
      def_delegator :@data, :each, :each_row
      def_delegator :@data, :each_with_index, :each_row_with_index
      def_delegator :@data, :[], :row
      def_delegator :@data, :size, :row_size
      def_delegators :@data, :first
      def value(row, column) @data[row][column] end
      def column_size(n=0)  (@data == []) ? 0 : @data[n].size end
      def column(n) @data.collect {|line| line[n]} end
      def subset(row_range)  self.class.new(row(row_range)) end
      def to_a()  row(0..-1)  end
    end
    
    class Scheme
      private
      class PositionedLabel
        def initialize(aNames, aPosition)
          @names = aNames #String[]
          @position = aPosition
          @note = {}
        end
        attr_reader :names, :position
        attr_accessor :note
      end
      public
      def initialize
        @body_position = nil
        @record_direction = nil
        @record_position = nil
        @record_size = nil
        @positioned_labels = []
      end
      attr_reader :positioned_labels
      attr_accessor :body_position, :record_direction, :record_position, :record_size
      def add_label(label_name, position)
        before_label = @positioned_labels.find{|l| l.position == position}
        if before_label
          before_label.names << label_name
        else
          @positioned_labels << PositionedLabel.new([label_name], position)
        end
      end
      def positioned_labels_to_labels
        @positioned_labels.collect {|pos_label| Label.new(pos_label.names, pos_label.note)}
      end
    end
    
    def initialize matrix
      @source = matrix
      @table = nil
      @scheme = nil
    end
    
    def read_table
      return nil  unless Tag::lookup_table_tag(@source.first)
      
      @scheme = Scheme.new
      title = read_table_header
      @table = Table.new(title)
      read_labels
      @table.records = read_records
      @table.labels = @scheme.positioned_labels_to_labels
      @table
    end
    
    private
    def read_table_header
      table_tag = Tag::lookup_table_tag(@source.first)
      raise "is not table"  unless table_tag
      title = table_tag[:data]
      @scheme.record_direction =  (table_tag[:tag] == Tag::HTABLE ? :horizontal : :vertical)
      @scheme.body_position = 1
      title
    end
    
    def read_labels
      @source.each_row_with_index do |values, row|
        next  if row < @scheme.body_position
        values.each_with_index do |value, column|
          read_labels_from value, row, column
        end
#        break  unless @record_position[:row] == row
      end
    end
    
    def read_labels_from(value, row, column)
      tagdata = Tag::lookup_tag(value, Tag::LABELS)
      return  unless tagdata
      
      case @scheme.record_direction
      when :vertical
        @scheme.record_position = row + 1
        @scheme.record_size = @source.column(column).size - @scheme.record_position
        label_names = [tagdata[:data]] + @source.row(row)[column+1 .. -1]
        label_start_position = column
        
      when :horizontal
        @scheme.record_position = column + 1
        @scheme.record_size = @source.row(row).size - @scheme.record_position
        label_names = [tagdata[:data]] + @source.column(column)[row+1 .. -1]
        label_start_position = row
      else
        raise
      end
      add_label label_names, label_start_position
    end
    
    def add_label label_names, label_start_position
      @previous_label_name = ""
      label_names.each_with_index do |label_name, i|
        label_name ||= ""
        if label_name != ""
          label_name = Tag.aliased(label_name)
        else
#          raise "x܂"  unless @previous_label_name
          label_name = @previous_label_name
        end
        @scheme.add_label label_name, label_start_position+i
        @previous_label_name = label_name
      end
    end
    
#    def directed_values(position, start_pos, end_pos)
#    end
    
    def key_position_of_label
      @scheme.positioned_labels.first.position
    end
    
    def read_records
      return []  if [@source, @scheme, @scheme.record_position].any? {|e| e == nil}
      record_last_position = @scheme.record_position + @scheme.record_size
      (@scheme.record_position ... record_last_position).collect do |pos|
        unless read_label_note_from pos
          read_record_from pos
        end
      end.compact
    rescue ArgumentError
      return []
    end
    
    def read_label_note_from pos
      note_tag = Tag::lookup_tag(read_value(key_position_of_label, pos), Tag::LABEL_NOTE)
      return nil  unless note_tag
      note_name = Tag.aliased(note_tag[:data])
      for pos_label in @scheme.positioned_labels do
        value = Tag.aliased read_value(pos_label.position, pos)
        pos_label.note[note_name] = value
      end
      true
    end
    
    def read_record_from pos
      record = Record.new
      for pos_label in @scheme.positioned_labels do
        value = Tag.aliased read_value(pos_label.position, pos)
        record.set pos_label.names, value
      end
      record
    end
    
    def read_value label_position, pos
      (@scheme.record_direction == :vertical) ? @source.value(pos, label_position) : @source.value(label_position,pos)
    end
    
    private
    def self.read_tables(any_format)
      matrix_list = any_format_to_matrixies any_format
      matrix_list.collect do |matrix|
        Reader.devide_into_table(matrix).collect do |table_lines|
          Reader.new(table_lines).read_table
        end
      end.flatten
    end
    
    def self.any_format_to_matrixies any_format
      case any_format
      when Array  #Array of Array of Value
        [ Matrix.new(any_format) ]
      when String
        case File.extname(any_format)
        when '.csv'
          [ Matrix.new(CSV.read(any_format)) ]
        when '.xls'
          matrixes_from_excel_file(any_format)
        else #String of CSV
          [ Matrix.new(CSV.parse(any_format)) ]
        end
      end
    end
  
    CHAR_CODE = 'sjis'
    def self.worksheet_to_matrix worksheet
      data = []
      worksheet.num_rows.times do |r|
        #        data << worksheet.row(r).collect {|cell| cell.to_s(CHAR_CODE)}
        data << worksheet.row(r).collect {|cell| cell.to_s == "" ? "" : cell.to_s(CHAR_CODE).to_s}
      end
      Matrix.new(data)
    end
  
    def self.matrixes_from_excel_file(file_path)
      result = []
      workbook = Spreadsheet::ParseExcel.parse(file_path)
      workbook.sheet_count.times do |sheet_num|
        result << worksheet_to_matrix(workbook.worksheet(sheet_num))
      end
      result
    end
    
    def self.start_of_table(line)
      Tag::lookup_table_tag(line) != nil
    end
    
    def self.end_of_table(line)
      (line.all? {|v| v == nil  or  v =~ /^\s*$/}) or start_of_table(line)
    end
    
    public
    def self.devide_into_table matrix
      result = []
      count = 0
      until (count >= matrix.row_size) do
        unless start_of_table( matrix.row(count) )
          count += 1
          next
        end
        start_of_table = count
        
        begin count += 1 end until (count >= matrix.row_size  or  Tag::lookup_tag(matrix.row(count),Tag::LABELS))
        count += 1  until  (count >= matrix.row_size  or end_of_table(matrix.row(count)) )
        result << matrix.subset(start_of_table...count)
      end
      result
    end
  end
end
